#!/usr/bin/env python2

#    This file is part of Anarcryst.
#    Anarcryst is copyright Andrew Walkingshaw <andrew@lexical.org.uk>, 2001- .

#    Anarcryst is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    Anarcryst is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Anarcryst; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import nested_scopes
import sys, os, random

errorstr = "" # global

class SectionError(Exception):
    """ Exception raised when the file has too many sections. """

    def __init__(self):
        pass


class MissingKeywordException(Exception):
    """ Exception raised when the input file lacks a keyword in the
        appropriate section. """
    
    def __init__(self):
        pass


class Atom:
    """ Storage class for a single atom in the basic unit cell. """
    
    def __init__(self, species, shell, locx, locy, locz):
        self._species = species
        self._locx = locx
        self._locy = locy
        self._locz = locz
        self._shell = shell
        
    def getDetails(self, unitX, unitY, unitZ, maxX, maxY, maxZ):
        """Return translated atom. """
        # now we want to be able to wrap things completely; if they
        # go out the cell, back to the start not to the local cell
        return (self._species, self._shell,
                ((float(self._locx) + int(unitX)) % maxX)/maxX,
                ((float(self._locy) + int(unitY)) % maxY)/maxY,
                ((float(self._locz) + int(unitZ)) % maxZ)/maxZ)


class AtomGroup:
    """ Storage class for a group of Atoms. """

    def __init__(self):
        self._weighting = 1
        self._atomlist = []
        self._vector = ""
        
    def setWeighting(self, weight):
        """ Set the probablistic weight of this group being used. """
        self._weighting = weight

    def setVector(self, str):
        self._vector = str

    def getVector(self):
        return self._vector
    
    def getWeighting(self):
        """ Get the probabilistic weighting of the likelihood of the current
        group being used. """
        return self._weighting

    def addAtom(self, atom):
        """ Add an atom to this group. """
        self._atomlist.append(atom)

    def getAtomList(self):
        """ Get the list of atoms in this group. """
        return self._atomlist

    def writeGroup(self, fd, cellx, celly, cellz, maxx, maxy, maxz):
        """ Output an AtomGroup to file. """
        for elem in self.getAtomList():
                    details = elem.getDetails(cellx, celly, cellz, maxx, maxy, maxz)
#                    print details
                    if details[1] == 1:
                        atype = "shel"
                    else:
                        atype = "core"
                        # construct the line for output...
                    ol = "%s %s %s %s %s\n" % (details[0], atype,
                                               details[2], details[3],
                                               details[4])
                    fd.write(ol)
class Parser:
    """ Class to hold and parse the input file format for anarcryst. """
    
    def __init__(self, infile):
        """ Open the input file and attempt to read it into a list.

        Fail if the file cannot be opened, exiting 2 on an OS error and 2 on
        an error relating to IO: non-existent files and/or permissions problems
        should raise an IOError leading to exiting with a return value of 1."""

        try:
            inputfd = open(infile, "r")
        except OSError:
            sys.stderr.write("OS Error; input file cannot be opened!")
            sys.exit(2)
        except IOError:
            sys.stderr.write(
                "Could not open file: does the input file exist,\
                and/or do you have sufficient permissions to access it?\n")
            sys.exit(1)
        self._input = inputfd.readlines()
        # and close the input file
        inputfd.close()
        self._header = []
        self._bulklines = []
        self._bulk = AtomGroup()
        self._disord = []
        self._disgrps = [[]]
        self._disweight = [0]
        self._footer = []
        self._keywords = []
        self._kw = {}
        
    def getHeader(self):
        return self._header
    
    def getBulk(self):
        return self._bulk

    def getFooter(self):
        return self._footer
    
    def _getDisgrps(self): # !warning! note plural
        return (self._disgrps, self._disweight) # !tuple!

    def getDisGrp(self):
        """ Select a disordered group from the parser object. """
        res = []
        disitems, totweights = self._getDisgrps()
        for slot in range(0, len(totweights)):
            totweight = totweights[slot]
            randtot = int(random.random() * int(totweight)) +1
            sys.stderr.write("%s %s\n" % (str(randtot), totweight))
            # should be an unbiased number 0 < x < totweight; therefore
            # with this algorithm should select groups with the correct
            # biases as per weighting
            for item in disitems[slot]:
                print item.getVector()
                # remove this item's weighting;
                randtot = randtot - item.getWeighting()
                sys.stderr.write("Remove weighting %s; " % item.getWeighting())
                sys.stderr.write("randtot now %s\n" % randtot)
                # and if we've exhausted the weight we generated, return
                # this item
                if randtot <= 0:
                    res.append(item)
                    break
                # if we get this far we're in trouble
            if randtot > 0:
                sys.stderr.write("Internal Error: failed to select a disgrp\n")
                sys.exit(-1)
        return res

    def getKw(self):
        return self._kw
    
    def parse(self):
        """ Split the opened input file into its sections.

        These should be:

        * header, with GULP keywords usw (self._header)

        * bulk structure: this is the same in all unit cells regardless
          of any disordering (self._bulk)

        * disordered input: this should be a list of positions of the atoms
          in the unit cell which take up a different position on ordering,
          with a weighting for each of these. (self._disord)

        * keywords:

          - number of repeated unit cells along x/y/z
          - number of output files to generate
          - root filenames
          - logfile name
          (self._keywords)

        Each of these will require accessor methods in the remainder of
        the class, but that should not be a problem.

        The section separator is '%%', by analogy with the fortune(1) file
        format. This must appear at the start of a line, and anything else
        on this line -will- be lost.

        If the file contains too many sections, failure with exit code 3
        will occur."""

        incr = 0 # temporary counter of what section we're in
        def append(line, incr):
            if line.startswith("#"): return # comment: skip the line
            if incr == 0: self._header.append(line)
            if incr == 1: self._bulklines.append(line)
            if incr == 2: self._disord.append(line)
            if incr == 3: self._footer.append(line)
            if incr == 4: self._keywords.append(line)
            if incr > 4: raise SectionError

        for line in self._input:
            if line.startswith("%%"):
                incr += 1
            else:
                try:
                    append(line, incr)
                except SectionError:
                    sys.stderr.write("Too many sections in input file!\n")
                    sys.exit(3)
                                     
        # and fall through on EOF when we run out of lines.
        # We should now have split the file up into sections,
        # but we have done precisely no sanity checking on the input.
        # We aren't going to do any on the header sections: those
        # get passed wholesale to the output file unchanged, so Not Our
        # Problem. Likewise with the footer.

        # Bulk needs parsing, though, to be able to translate it to
        # new unit cells, which is irritating.
        # The parsing routine from the randomised structure suitably
        # simplified should work:

        for line in self._bulklines:
                thisline = line.split()
                if line.lstrip() == "":
                    continue
                elif thisline[1].startswith("shel"):
                    shell = 1
                else:
                    shell = 0
                # set up an atom with these parameters
                thisAtom = Atom(thisline[0], shell, thisline[2],
                                thisline[3], thisline[4])
                # and attach it to the current group.
                self._bulk.addAtom(thisAtom)
        # Keywords (problems here bomb returning exit code 4):
        try:
            for line in self._keywords:
                if line.startswith("numx"):
                    self._kw["numx"] = int(line.split(":")[1])
                elif line.startswith("numy"):
                    self._kw["numy"] = int(line.split(":")[1])
                elif line.startswith("numz"):
                    self._kw["numz"] = int(line.split(":")[1])
                elif line.startswith("numoutputs"):
                    self._kw["numoutputs"] = int(line.split(":")[1])
                elif line.startswith("rootfilename"):
                    # this is a string, so don't test it by casting to int
                    self._kw["rootfilename"] = line.split(":")[1]
                elif line.startswith("logfilename"):
                    self._kw["logfilename"] = line.split(":")[1]
        except ValueError: # invalid cast to int
            sys.stderr.write("Invalid argument to numeric keyword\n")
            sys.exit(4)

        # Now test to see if the dictionary has all the resulting keys;
        try:
            for key in ("numx", "numy", "numz", "numoutputs", "rootfilename",
                        "logfilename"):
                if not self._kw.has_key(key):
                    errorstr = key
                    raise MissingKeywordException()
        except MissingKeywordException:
            sys.stderr.write("Missing keyword: %s\n" % errorstr)
            sys.exit(4)

        # end keywords
        # Parsing the possible bits of the structure is... um...
        # "rather a lot harder". There are two plausible approaches; one is
        # to just take gobbets of text, but we know this can't work because
        # we need to be able to translate it at will to other unit cell
        # positions. Hence, we need to do this properly and define a parser:
        # %%
        # ...
        # ++ 
        # $$
        # species type x-coord y-coord z-coord
        # species type x-coord y-coord z-coord
        # species type x-coord y-coord z-coord
        # species type x-coord y-coord z-coord
        # weighting: value
        # vector: string
        # $$
        # ...
        # ++
        # $$
        # ...
        # %%
        # per block being a relatively sensible way of delimiting these.
        # 
        # start disord
        curindx = 0
        curslot = 0
        self._disgrps[curslot].append(AtomGroup()) # !inline instantiation!
        curgrp = self._disgrps[curslot][curindx]
        for line in self._disord:
            if line.startswith("++"): # slot delimiter
                self._disgrps.append([AtomGroup()])
                # add a new slot inlining instantiation
                self._disweight.append(0) # and weighting sum
                curslot += 1 # move to it
                curindx = 0  # and reset to the start of this slot
                curgrp = self._disgrps[curslot][curindx]
            elif line.startswith("$$"): # block delimiter
                # create a new group
                self._disgrps[curslot].append(AtomGroup())
                curindx += 1
                print curslot
                print curindx
                print self._disgrps
                curgrp = self._disgrps[curslot][curindx]
            elif line.startswith("vector"):
                curgrp.setVector(line.split(":")[1])
            elif line.startswith("weighting"):
                try:
                    # cast to catch the obvious error
                    curgrp.setWeighting(int(line.split(":")[1]))
                    self._disweight[curslot] += curgrp.getWeighting()
                except ValueError:
                    sys.stderr.write("Non-integer argument to 'weighting:'.\n")
                    sys.exit(5)
            elif len(line.lstrip()) == 0:
                continue
            else:
                thisline = line.split()
                shell = 0
                if thisline[1].startswith("shel"): shell = 1
                # set up an atom with these parameters
                thisAtom = Atom(thisline[0], shell, thisline[2],
                                thisline[3], thisline[4])
                # and attach it to the current group.
                curgrp.addAtom(thisAtom)
        del curindx
